/* crash-pipe: handle core file passed from stdin
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author: Karol Lewandowski <k.lewandowsk@samsung.com>
 */
#define _GNU_SOURCE 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <syslog.h>
#include <sys/prctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <limits.h>
#include <elf.h>
#include <signal.h>
#include <sys/procfs.h>

#include "crash-pipe-arch.h"

#define NELEMS(arr) (sizeof(arr)/sizeof(arr[0]))
#define BUF_SIZE (BUFSIZ)
#define HEXA 16
#define PERM_LEN 5
#define ADDR_LEN 16
#define STR_ANONY "[anony]"
#define STR_ANONY_LEN 8

#define roundup(x,y) (				\
{						\
	(((x) + (y-1))/y) * y;			\
}						\
)

enum {
	OPT_HELP,
	OPT_REPORT,
	OPT_SAVE_CORE,
};

struct file_mapping {
	long start;
	long end;
	long file_ofs;
};

typedef struct {
	long count;
	long page_size;
	struct file_mapping mappings[0];
} mapped_files_t;

struct addr_node {
	long *startaddr;
	long *endaddr;
	char perm[5];
	char *fpath;
	struct addr_node *next;
};

const struct option opts[] = {
	{ "help", no_argument, 0, OPT_HELP },
	{ "report", no_argument, 0, OPT_REPORT },
	{ "save-core", required_argument, 0, OPT_SAVE_CORE },
	{ 0, 0, 0, 0 }
};

static char *argv0 = "";

static int core_input;
static int core_output;
static char* core_buffer;
static int core_bytes;
#define CORE_BUFFER_SIZE 16384

static unsigned long sp;
static unsigned long stack_start;
static unsigned long stack_end;

static void usage(void)
{
	fprintf(stderr, "usage: %s [--help] [--save-core FILE_NAME] [--report] PID UID GID SIGNAL DUMPTIME EXE\n",
			argv0);
}

/* read file to buffer
 *
 * Returns success only if whole file has been read (requires big
 * enough buffer).
 */
static int procfs_read_fileline(const char *pid, const char *filename, char *outbuf, int outsize)
{
	char *path = NULL;
	int fd;
	int n;
	int ret = 0;

	if (!(outsize > 0))
		return 0;

	if (asprintf(&path, "/proc/%s/%s", pid, filename) == -1)
		return -ENOMEM;

	fd = open(path, O_RDONLY);
	if (fd == -1) {
		ret = -errno;
		goto err;
	}

	/* XXX we are really assuming here that one read is enough */
	ret = read(fd, outbuf, outsize);
	if (ret == -1 || ret == outsize /* no place for \0 */) {
		ret = -errno;
		goto err;
	}

	n = ret;
	outbuf[n] = 0;
	for (; n > 0; --n) {
		if (outbuf[n] == '\n')
			outbuf[n] = 0;
	}

	close(fd);

	free(path);
	return ret;

err:
	if (fd >= 0)
		close(fd);
	free(path);
	*outbuf = 0;
	return ret;
}

void print_multiline(char *buf, int buf_size)
{
	int i;
	int pos;

	for (pos = i = 0; buf[pos] && pos < buf_size; ++i, pos += strlen(buf + pos) + 1)
		printf("%21d: %s\n", i, buf + pos);
}

static char *fgets_fd(char *str, int len, int fd)
{
        char ch;
        register char *cs;
        int num = 0;

        cs = str;
        while (--len > 0 && (num = read(fd, &ch, 1) > 0)) {
                if ((*cs++ = ch) == '\n')
                        break;
        }
        *cs = '\0';
        return (num == 0 && cs == str) ? NULL : str;
}

static void meminfo_report(const char *pidstr)
{
	char infoname[BUF_SIZE];
	char memsize[BUF_SIZE];
	char linebuf[BUF_SIZE];
	char file_path[PATH_MAX];
	int fd;

	printf("\nMemory information\n");

        if ((fd = open("/proc/meminfo", O_RDONLY)) < 0) {
                fprintf(stderr, "[crash-pipe] can't open %s\n", file_path);
        } else {
                while (fgets_fd(linebuf, BUF_SIZE, fd) != NULL) {
                        sscanf(linebuf, "%s %s %*s", infoname, memsize);
                        if (strcmp("MemTotal:", infoname) == 0) {
                                printf("%s %8s KB\n", infoname, memsize);
                        } else if (strcmp("MemFree:", infoname) == 0) {
                                printf("%s  %8s KB\n", infoname, memsize);
                        } else if (strcmp("Buffers:", infoname) == 0) {
                                printf("%s  %8s KB\n", infoname, memsize);
                        } else if (strcmp("Cached:", infoname) == 0) {
                                printf("%s   %8s KB\n", infoname, memsize);
                                break;
                        }
                }
                close(fd);
        }

	snprintf(file_path, PATH_MAX, "/proc/%s/status", pidstr);
        if ((fd = open(file_path, O_RDONLY)) < 0) {
                fprintf(stderr, "[crash-pipe]can't open /proc/meminfo\n");
        } else {
                while (fgets_fd(linebuf, BUF_SIZE, fd) != NULL) {
                        sscanf(linebuf, "%s %s %*s", infoname, memsize);
                        if (strcmp("VmPeak:", infoname) == 0) {
                                printf("%s   %8s KB\n", infoname,
                                                memsize);
                        } else if (strcmp("VmSize:", infoname) == 0) {
                                printf("%s   %8s KB\n", infoname,
                                                memsize);
                        } else if (strcmp("VmLck:", infoname) == 0) {
                                printf("%s    %8s KB\n", infoname,
                                                memsize);
                        } else if (strcmp("VmPin:", infoname) == 0) {
                                printf("%s    %8s KB\n", infoname,
                                                memsize);
                        } else if (strcmp("VmHWM:", infoname) == 0) {
                                printf("%s    %8s KB\n",
                                                infoname, memsize);
                        } else if (strcmp("VmRSS:", infoname) == 0) {
                                printf("%s    %8s KB\n",
                                                infoname, memsize);
                        } else if (strcmp("VmData:", infoname) == 0) {
                                printf("%s   %8s KB\n",
                                                infoname, memsize);
                        } else if (strcmp("VmStk:", infoname) == 0) {
                                printf("%s    %8s KB\n",
                                                infoname, memsize);
                        } else if (strcmp("VmExe:", infoname) == 0) {
                                printf("%s    %8s KB\n",
                                                infoname, memsize);
                        } else if (strcmp("VmLib:", infoname) == 0) {
                                printf("%s    %8s KB\n",
                                                infoname, memsize);
                        } else if (strcmp("VmPTE:", infoname) == 0) {
                                printf("%s    %8s KB\n",
                                                infoname, memsize);
                        } else if (strcmp("VmSwap:", infoname) == 0) {
                                printf("%s   %8s KB\n",
                                                infoname, memsize);
                                break;
                        }
                }
                close(fd);
        }
}

static struct addr_node *get_addr_list_from_maps(int fd)
{
        int fpath_len, result;
        long *saddr;
        long *eaddr;
        char perm[PERM_LEN];
        char path[PATH_MAX];
        char addr[ADDR_LEN * 2];
        char linebuf[BUF_SIZE];
        struct addr_node *head = NULL;
        struct addr_node *tail = NULL;
        struct addr_node *t_node = NULL;

        /* parsing the maps to get executable code address */
        while (fgets_fd(linebuf, BUF_SIZE, fd) != NULL) {
                memset(path, 0, PATH_MAX);
                result = sscanf(linebuf, "%s %s %*s %*s %*s %s ", addr, perm, path);
                if (result < 0)
                        continue;
                perm[PERM_LEN - 1] = 0;
                /* rwxp */
                if ((perm[2] == 'x' && path[0] == '/') ||
		    (perm[1] == 'w' && path[0] != '/'))
		{
			/* add addr node to list */
			addr[ADDR_LEN] = 0;
			saddr = (long *)strtoul(addr, NULL, HEXA);
			/* ffff0000-ffff1000 */
			eaddr = (long *)strtoul(&addr[ADDR_LEN + 1], NULL, HEXA);
			/* make node and attach to the list */
			t_node = (struct addr_node *)mmap(0, sizeof(struct addr_node),
							  PROT_READ | PROT_WRITE,
							  MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
			if (t_node == NULL) {
				fprintf(stderr, "error : mmap\n");
				return NULL;
			}
			memcpy(t_node->perm, perm, PERM_LEN);
			t_node->startaddr = saddr;
			t_node->endaddr = eaddr;
			t_node->fpath = NULL;
			fpath_len = strlen(path);
			if (fpath_len > 0) {
				t_node->fpath = (char *)mmap(0, fpath_len + 1,
							     PROT_READ | PROT_WRITE,
							     MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
				memset(t_node->fpath, 0, fpath_len + 1);
				memcpy(t_node->fpath, path, fpath_len);
			} else {
				t_node->fpath = (char *)mmap(0, STR_ANONY_LEN,
							     PROT_READ | PROT_WRITE,
							     MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
				memset(t_node->fpath, 0, STR_ANONY_LEN);
				memcpy(t_node->fpath, STR_ANONY, STR_ANONY_LEN);
			}
			t_node->next = NULL;
			if (head == NULL) {
				head = t_node;
				tail = t_node;
			} else {
				tail->next = t_node;
				tail = t_node;
			}
		}
	}
	return head;
}

static void free_all_nodes(struct addr_node *start)
{
	struct addr_node *t_node, *n_node;
	int fpath_len;

	if (start == NULL)
		return;
	t_node = start;
	n_node = t_node->next;
	while (t_node) {
		if (t_node->fpath != NULL) {
			fpath_len = strlen(t_node->fpath);
			munmap(t_node->fpath, fpath_len + 1);
		}
		munmap(t_node, sizeof(struct addr_node));
		if (n_node == NULL)
			break;
		t_node = n_node;
		n_node = n_node->next;
	}
}

static void print_node_to_file(struct addr_node *start)
{
	struct addr_node *t_node;

	t_node = start;
	printf("\n%s\n", "Maps Information");
	while (t_node) {
		if (!strncmp(STR_ANONY, t_node->fpath, STR_ANONY_LEN)) {
			t_node = t_node->next;
		} else {
			printf( "%16lx %16lx %s %s\n",
				(unsigned long)t_node->startaddr,
				(unsigned long)t_node->endaddr,
				t_node->perm, t_node->fpath);
			t_node = t_node->next;
		}
	}
	printf("%s\n", "End of Maps Information");
}

static void maps_report(const char *pidstr)
{
	char file_path[PATH_MAX];
	struct addr_node *head = NULL;
	int fd;

	/* open maps file */
	snprintf(file_path, PATH_MAX, "/proc/%s/maps", pidstr);

        if ((fd = open(file_path, O_RDONLY)) < 0) {
                fprintf(stderr, "[crash-pipe]can't open %s\n", file_path);
        } else {
                /* parsing the maps to get code segment address*/
                head = get_addr_list_from_maps(fd);
                close(fd);
        }
	if (head != NULL) {
		/* print maps information */
		print_node_to_file(head);
		free_all_nodes(head);
	}
}

static void report(int argc, char *argv[])
{
	const char *pidstr = argv[0];
	const char *uidstr = argv[1];
	const char *gidstr = argv[2];
	const char *sigstr = argv[3];
	const char *timestr = argv[4];
	const char *exestr = argv[5];
	static const struct {
		char *file;
		char *desc;
		int is_multiline;
	} proc_filedesc[] = {
		{ "comm", "Comm", 0},
		{ "cgroup", "CGroup", 0 },
		{ "attr/current", "MAC Label", 0 },
		{ "oom_score", "OOM Score", 0 },
		{ "oom_score_adj", "OOM Score Adj", 0 },
		{ "cmdline", "Cmdline", 1 },
		{ "environ", "Environment", 1 }
	};

	int i;
	int n;

#define PROC_READ_MAX 16384 /* 4 pages should be enough for any process */
	char proc_readbuf[PROC_READ_MAX];
	char exe_link[PATH_MAX];
	char exe_file[PATH_MAX];

	snprintf(exe_link, PATH_MAX, "/proc/%s/exe", pidstr);
	if (readlink(exe_link, exe_file, PATH_MAX) > 0) {
		printf("Executable File Path: %s\n", exe_file);
	}

	meminfo_report(pidstr);

	maps_report(pidstr);

	printf(" - passed from kernel -\n"
		   "%16s: %s\n"
		   "%16s: %s\n"
		   "%16s: %s\n"
		   "%16s: %s\n"
		   "%16s: %s\n"
		   "%16s: %s\n\n",
		   "PID", pidstr,
		   "UID", uidstr,
		   "GID", gidstr,
		   "Signal number", sigstr,
		   "Timestamp", timestr,
		   "Executable", exestr);

	printf(" - procfs information -\n");

	for (i = 0; i < NELEMS(proc_filedesc); ++i) {
		n = procfs_read_fileline(pidstr, proc_filedesc[i].file, proc_readbuf, sizeof(proc_readbuf));
		if (n < 0)
			snprintf(proc_readbuf, sizeof(proc_readbuf), "Error (%d)", -n);

		if (n < 0 || proc_filedesc[i].is_multiline == 0)
			printf("%16s: %s\n", proc_filedesc[i].desc, proc_readbuf);
		else {
			printf("%16s:\n", proc_filedesc[i].desc);
			print_multiline(proc_readbuf, n);
		}
	}
}

static int open_core(const char *core_path)
{
	int fd;
	fd = open(core_path, O_WRONLY | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
	if (fd == -1) {
		syslog(LOG_ERR, "crash-pipe: Unable to save core file to %s: %m\n", core_path);
		return -errno;
	}

	return fd;
}

static int save_core(int fd, const char* header, int headersz)
{
	char *buf[4096];
	int readb, remaining;
	int ret = 0;
	int n;

	for (n = 0, remaining = headersz ; remaining > 0; remaining -= n) {
		n = write(fd, header, remaining);
		if (n == -1) {
			ret = -errno;
			syslog(LOG_ERR, "crash-pipe:%d: Error while saving core file: %m.\n", __LINE__);
			return ret;
		}
	}

	while ((readb = read(core_input, buf, sizeof(buf))) > 0) {
		for (n = 0, remaining = readb ; remaining > 0; remaining -= n) {
			n = write(fd, buf, remaining);
			if (n == -1) {
				ret = -errno;
				syslog(LOG_ERR, "crash-pipe:%d: Error while saving core file: %m.\n", __LINE__);
				return ret;
			}
		}
	}

	return ret;
}


static int read_n_bytes_of_core(char* b, int remaining)
{
	int ret=0;
	int readb;
	while(remaining > 0) {
		readb = read(core_input, b, remaining);
		if (readb < 0){
			syslog(LOG_ERR, "crash-pipe:%d: Error reading core: %m.", __LINE__);
			return readb;
		}
		remaining -= readb;
		b += readb;
		ret += readb;
	}
	core_bytes += ret;
	return ret;
}

static void dump_siginfo(siginfo_t *si)
{
	const char* const signal_table[] = {
		[SIGHUP]="SIGHUP", [SIGINT]="SIGINT", [SIGQUIT]="SIGQUIT",
		[SIGILL]="SIGILL", [SIGTRAP]="SIGTRAP", [SIGABRT]="SIGABRT",
		[SIGIOT]="SIGIOT", [SIGBUS]="SIGBUS", [SIGFPE]="SIGFPE",
		[SIGKILL]="SIGKILL", [SIGUSR1]="SIGUSR1", [SIGSEGV]="SIGSEGV",
		[SIGUSR2]="SIGUSR2", [SIGPIPE]="SIGPIPE", [SIGALRM]="SIGALRM",
		[SIGTERM]="SIGTERM", [SIGSTKFLT]="SIGSTKFLT", [SIGCHLD]="SIGCHLD",
		[SIGCONT]="SIGCONT", [SIGSTOP]="SIGSTOP", [SIGTSTP]="SIGTSTP",
		[SIGTTIN]="SIGTTIN", [SIGTTOU]="SIGTTOU", [SIGURG]="SIGURG",
		[SIGXCPU]="SIGXCPU", [SIGXFSZ]="SIGXFSZ", [SIGVTALRM]="SIGVTALRM",
		[SIGPROF]="SIGPROF", [SIGWINCH]="SIGWINCH", [SIGIO]="SIGIO",
		[SIGPWR]="SIGPWR", [SIGSYS]="SIGSYS", [SIGUNUSED]="SIGUNUSED",
	};
	printf("Signal: %d\n"
	       "\t(%s)\n"
	       "\tsi_code: %d\n",
	       si->si_signo,
	       signal_table[si->si_signo],
	       si->si_code);
	switch (si->si_code) {
	case SI_TKILL:
	case SI_USER:
		printf("\tsignal sent by %s (sent by pid %d, uid %d)",
		       si->si_code == SI_TKILL ? "tkill" : "kill",
		       si->si_pid, si->si_uid);
		break;
	case SI_KERNEL:
		printf("\tsignal sent by the kernel\n");
		break;
	}
	printf("\n");
}

static void dump_auxv(elf_addr_t *elf_info)
{
	int i;
	const char* const at_table[] = {
		[AT_IGNORE] = "IGNORE", [AT_EXECFD] = "EXECFD",
		[AT_PHDR] = "PHDR", [AT_PHENT] = "PHENT",
		[AT_PHNUM] = "PHNUM", [AT_PAGESZ] = "PAGESZ",
		[AT_BASE] = "BASE", [AT_FLAGS] = "FLAGS",
		[AT_ENTRY] = "ENTRY", [AT_NOTELF] = "NOTELF",
		[AT_UID] = "UID", [AT_EUID] = "EUID",
		[AT_GID] = "GID", [AT_EGID] = "EGID",
		[AT_CLKTCK] = "CLKTCK", [AT_PLATFORM] = "PLATFORM",
		[AT_HWCAP] = "HWCAP", [AT_FPUCW] = "FPUCW",
		[AT_DCACHEBSIZE] = "DCACHEBSIZE",
		[AT_ICACHEBSIZE] = "ICACHEBSIZE",
		[AT_UCACHEBSIZE] = "UCACHEBSIZE",
		[AT_IGNOREPPC] = "IGNOREPPC",
		[AT_SECURE] = "SECURE",
		[AT_BASE_PLATFORM] = "BASE_PLATFORM",
		[AT_RANDOM] = "RANDOM", [AT_HWCAP2] = "HWCAP2",
		[AT_EXECFN] = "EXECFN", [AT_SYSINFO] = "SYSINFO",
		[AT_SYSINFO_EHDR] = "SYSINFO_EHDR",
		[AT_L1I_CACHESHAPE] = "L1I_CACHESHAPE",
		[AT_L1D_CACHESHAPE] = "L1D_CACHESHAPE",
		[AT_L2_CACHESHAPE] = "L2_CACHESHAPE",
		[AT_L3_CACHESHAPE] = "L3_CACHESHAPE"};

	printf("Auxiliary vector:\n");
	for (i = 0; elf_info[i] != AT_NULL;) {
		elf_addr_t k = elf_info[i++];
		elf_addr_t v = elf_info[i++];
		switch (k) {
		case AT_PAGESZ:
		case AT_CLKTCK:
		case AT_PHENT:
		case AT_PHNUM:
		case AT_FLAGS:
		case AT_UID:
		case AT_EUID:
		case AT_GID:
		case AT_EGID:
		case AT_SECURE:
			printf("\t%-14s: %d\n", at_table[k], v);
			break;
		default:
			printf("\t%-14s: 0x%x\n", at_table[k], v);
		}
	}
	printf("\n");
}

static void dump_files(mapped_files_t *mf)
{
	long i;
	char *fn = (char*)&mf->mappings[mf->count];
	printf("Mapped files:\n");

	for (i=0; i < mf->count; i++) {
		printf("%lx-%lx %lx %s\n",
		       mf->mappings[i].start,
		       mf->mappings[i].end,
		       mf->mappings[i].file_ofs * mf->page_size,
		       fn);
		fn += strlen(fn) + 1;
	}

	printf("\n");
}

static void dump_prstatus(struct elf_prstatus* prstatus)
{
	arch_dump_registers(&prstatus->pr_reg);
	printf("\n");
}

static int parse_core()
{
	int i;
	char *b;
	elf_ehdr *eh;
	elf_phdr *ph, *phi;
	elf_nhdr *nh, *nhi;
	int readb;

	eh = (void*)(b = core_buffer);

	readb = read_n_bytes_of_core(b, sizeof(*eh));
	if (readb < 0){
		syslog(LOG_ERR, "crash-pipe:%d: Error reading core: %m.", __LINE__);
		return readb;
	}
	b += readb;

	if (memcmp(eh->e_ident, ELFMAG, 4) != 0 ||
	    eh->e_ident[EI_CLASS] != _ELFCLASS ||
	    (eh->e_ident[EI_OSABI] != ELFOSABI_LINUX &&
	     eh->e_ident[EI_OSABI] != ELFOSABI_NONE) ||
	    eh->e_type != ET_CORE) {
		syslog(LOG_ERR, "crash-pipe:%d: Unsupported core format.", __LINE__);
		return -1;
	}

	/* Read everything up to programme headers */
	readb = read_n_bytes_of_core(b, eh->e_phoff - (b - core_buffer));
	if (readb < 0){
		syslog(LOG_ERR, "crash-pipe:%d: Error reading core: %m.", __LINE__);
		return readb;
	}
	b += readb;

	readb = read_n_bytes_of_core(b, eh->e_phentsize * eh->e_phnum);
	if (readb < 0) {
		syslog(LOG_ERR, "crash-pipe:%d: Error reading core: %m.", __LINE__);
		return readb;
	}
	ph = (elf_phdr*)b;
	b += readb;

	for (nh = NULL, phi = ph, i=0; i < eh->e_phnum; i++, phi++) {
		if (phi->p_type == PT_NOTE) {
			readb = read_n_bytes_of_core(b,
						     phi->p_offset - (b - core_buffer) +
						     phi->p_filesz);
			if (read < 0){
				syslog(LOG_ERR, "crash-pipe:%d: Error reading core: %m.", __LINE__);
				return readb;
			}
			nh = (elf_nhdr*)b;
			b += readb;
			break;
		}
	}

	if (nh  == NULL) {
		syslog(LOG_ERR, "crash-pipe:%d: Note programme header not found.", __LINE__);
		return -1;
	}

	for (nhi = nh;
	     (uintptr_t)nhi < (uintptr_t)(core_buffer + ph->p_offset + ph->p_filesz);
	     nhi = (elf_nhdr*)((uintptr_t)nhi + sizeof(*nh) +
			       roundup(nhi->n_namesz, 4) +
			       roundup(nhi->n_descsz, 4))) {
		void *data = (void*)((uintptr_t)nhi + sizeof(*nh) +
				     roundup(nhi->n_namesz, 4));

		switch (nhi->n_type) {
		case NT_SIGINFO:
			dump_siginfo((siginfo_t*)data);
			break;
		case NT_FILE:
			dump_files((mapped_files_t*)data);
			break;
		case NT_AUXV:
			dump_auxv((elf_addr_t*)data);
			break;
		case NT_PRSTATUS:
			dump_prstatus((struct elf_prstatus*)data);
			sp = arch_get_sp(&(((struct elf_prstatus*)data)->pr_reg));
			break;
		case NT_FPREGSET:
		case NT_PRPSINFO:
		case NT_X86_XSTATE:
		default:
			printf("offset: 0x%08x note type: 0x%08x\n",
			       ((uintptr_t)nhi - (uintptr_t)core_buffer),
			       nhi->n_type);
		}
	}

	for (phi = ph, i=0; i < eh->e_phnum; i++, phi++) {
		if (phi->p_flags == (PF_R | PF_W) &&
		    phi->p_vaddr < sp &&
		    phi->p_vaddr + phi->p_memsz > sp) {
			stack_start = phi->p_vaddr;
			stack_end =  phi->p_vaddr + phi->p_memsz;
			printf("Stack location: ");
			printf("%lx-%lx\n", stack_start, stack_end);

		}
	}

	return readb;
}

int main(int argc, char *argv[])
{
	int c;
	int opt_report = 0;
	char *opt_save_core = NULL;
	int ret = 1;

	prctl(PR_SET_DUMPABLE, 0);

	argv0 = argv[0];

	core_input = STDIN_FILENO;

	while ((c = getopt_long_only(argc, argv, "", opts, NULL)) != -1) {

		if (c == OPT_HELP) {
			usage();
			exit(EXIT_SUCCESS);
		} else if (c == OPT_REPORT) {
			opt_report = 1;
		} else if (c == OPT_SAVE_CORE) {
			opt_save_core = strdup(optarg);
			if (!opt_save_core) {
				syslog(LOG_CRIT, "Out of memory. Exiting.");
				exit(EXIT_FAILURE);
			}
		}
	}

	argc -= optind;
	argv += optind;

	core_buffer = malloc(CORE_BUFFER_SIZE);
	if (core_buffer == NULL) {
		syslog(LOG_ERR, "crash-pipe:%d: Unable to allocate memory: %m\n", __LINE__);
		return EXIT_FAILURE;
	}
	memset(core_buffer, 0, CORE_BUFFER_SIZE);

	parse_core();

	if (opt_report)
		report(argc, argv);

	if (opt_save_core) {
		core_output = open_core(opt_save_core);
		if (core_output < 0);
		ret = save_core(core_output, core_buffer, core_bytes);
		close(core_output);
		if (ret < 0)
			unlink(opt_save_core);
	}

	free(core_buffer);
	return ret >= 0 ? EXIT_SUCCESS : EXIT_FAILURE;
}
