/*
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author: Adrian Szyndela <adrian.s@samsung.com>
 */
/*
 * @file crash-stack.h
 * @brief unwinding call stacks, header file for common functions
 */
#ifndef CRASH_STACK_H
#define CRASH_STACK_H

#include <stdint.h>
#include <elfutils/libdwfl.h>

#define MAX_CALLSTACK_LEN 1000

/**
 * @brief call stack database
 */
struct Callstack {
	uintptr_t tab[MAX_CALLSTACK_LEN]; 	///< storage for elements
	size_t elems;						///< number of elements in the database
};
typedef struct Callstack Callstack;

/**
 * @brief module mappings database element - a single mapping
 */
struct Mapping {
	uintptr_t m_start;					///< start address of the module
	uintptr_t m_end;					///< end address of the module
	uintptr_t m_offset;					///< offset in core file
	const char *m_name;					///< file name of the module
	int m_fd;							///< file descriptor, filled if already open
	Elf *m_elf;							///< ELF handler, initialized when ELF open with libelf
};
typedef struct Mapping Mapping;

#define MAX_MAPPINGS_NUM 1000
/**
 * @brief module mappings database
 */
struct Mappings {
	Mapping tab[MAX_MAPPINGS_NUM];		///< storage for elements
	size_t elems;						///< number of elements in the database
};
typedef struct Mappings Mappings;

/**
 * @brief Gets address of storage for specified register value
 *
 * @remarks This function uses static storage. Thus, it shouldn't be used in multiple threads.
 *
 * @param regname name of the register, e.g. "pc", "eax", "r7"
 * @param regnum ordinary number of the register, when name is not available
 * @return address of statically allocated storage of proper size for the register value
 */
void *_get_place_for_register_value(const char *regname, int regnum);

/**
 * @brief Creates call stack database.
 *
 * @param dwfl dwfl handle
 * @param core core file handle, NULL if live process is analyzed
 * @param pid pid of the analyzed process, 0 if core dump file is analyzed
 * @param mappings module mappings database
 * @param callstack pointer to call stack database; results are put here
 */
void _create_crash_stack(Dwfl *dwfl, Elf *core, pid_t pid, Mappings *mappings, Callstack *callstack);

/**
 * @brief Tries to find address of the prologue of the function specified by some address.
 *
 * @param dwfl dwfl handle
 * @param current_pc some address from a function
 * @param mappings module mappings database
 * @return address of the found prologue of the function, NULL if not found
 */
Dwarf_Addr _crash_stack_libelf_get_prologue_pc(Dwfl *dwfl, Dwarf_Addr current_pc, Mappings *mappings);

/**
 * @brief Reads a value from specified address from target's address space.
 *
 * @param dwfl dwfl handle
 * @param core core file handle, NULL if live process is analyzed
 * @param pid pid of the analyzed process, 0 if core dump file is analyzed
 * @param a address to read from
 * @param[out] v allocated memory to copy value to
 * @param size size of the value to copy
 * @param mappings module mappings database
 * @return true on success, false otherwise
 */
bool _crash_stack_libelf_read_value(Dwfl *dwfl, Elf *core, pid_t pid,
									Dwarf_Addr a, void *v, size_t size,
									Mappings *mappings);

/**
 * @brief Gets allocated space for getting target registers with ptrace().
 *
 * @remarks This function uses static storage. Thus, it shouldn't be used in multiple threads.
 *
 * @param[out] size size of the registers structure
 * @return address of statically allocated storage of proper size for the registers
 */
void *_crash_stack_get_memory_for_ptrace_registers(size_t *size);

/**
 * @brief Sets current values of registers from the ptrace buffer
 *
 * @param regbuf buffer used for getting registers with ptrace()
 *
 * @see _crash_stack_get_memory_for_ptrace_registers()
 */
void _crash_stack_set_ptrace_registers(void *regbuf);

/**
 * @brief Print values of registers from the ptrace buffer
 *
 * @param outputfile output stream
 */
void _crash_stack_print_regs(FILE* outputfile);
#endif /* CRASH_STACK_H */
